import sys
from enum import Enum
from pathlib import Path
from datetime import datetime, timedelta
from typing import Union, List, Optional
import pandas as pd
import modin.pandas as mpd
import numpy as np
import numbers
import plotly
import plotly.express as px

from cached_property import cached_property

from .files import cached_download_cpi
from .dates import convert_date
from .viz import format_fig


class Location(str, Enum):
    AUSTRALIA = "Australia"
    SYDNEY = "Sydney"
    MELBOURNE = "Melbourne"
    BRISBANE = "Brisbane"
    ADELAIDE = "Adelaide"
    PERTH = "Perth"
    HOBART = "Hobart"
    DARWIN = "Darwin"
    CANBERRA = "Canberra"

    def __str__(self):
        return self.value


class CPI:
    """A class to manage the Australian Consumer Index (CPI) data."""

    @cached_property
    def latest_cpi_df(self) -> pd.DataFrame:
        """
        Returns a Pandas DataFrame with the latest Australian Consumer Price Index (CPI) data.

        Returns:
            pd.DataFrame: The latest Australian Consumer Price Index (CPI) data. The index of the series is the relevant date for each row.
        """
        local_path = cached_download_cpi()
        excel_file = pd.ExcelFile(local_path)
        df = excel_file.parse("Data1")

        # Get rid of extra headers
        df = df.iloc[9:]

        df = df.rename(columns={"Unnamed: 0": "Date"})
        df.index = df["Date"]

        return df

    def column_name(self, location: Union[Location, str] = Location.AUSTRALIA):
        return f"Index Numbers ;  All groups CPI ;  {str(location).title()} ;"

    def cpi_series(self, location: Union[Location, str] = Location.AUSTRALIA) -> pd.Series:
        """
        Returns a Pandas Series with the Australian CPI (Consumer Price Index) per quarter.

        This is taken from the latest spreadsheet from the Australian Bureau of Statistics (file id '640101').
        The index of the series is the relevant date.

        Args:
            location (Union[Location, str], optional): The location for calculating the CPI.
                Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
                Default is 'Australia'.

        Returns:
            pd.Series: The CPIs per quarter.
        """
        df = self.latest_cpi_df

        return df[self.column_name(location)]

    def cpi_at(
        self, date: Union[datetime, str, pd.Series, np.ndarray], location: Union[Location, str] = Location.AUSTRALIA
    ) -> Union[float, np.ndarray]:
        """
        Returns the CPI (Consumer Price Index) for a date (or a number of dates).

        If `date` is a string then it is converted to a datetime using dateutil.parser.
        If `date` is a vector then it returns a vector otherwise it returns a single scalar value.
        If `date` is before the earliest reference date (i.e. 1948-09-01) then it returns a NaN.

        Args:
            date (Union[datetime, str, pd.Series, np.ndarray]): The date(s) to get the CPI(s) for.
            location (Union[Location, str], optional): The location for calculating the CPI.
                Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
                Default is 'Australia'.

        Returns:
            Union[float, np.ndarray]: The CPI value(s).
        """
        date = convert_date(date)

        cpi_series = self.cpi_series(location=location)

        min_date = cpi_series.index.min()
        cpis = np.array(
            cpi_series[np.searchsorted(cpi_series.index, date, side="right") - 1],
            dtype=float,
        )
        cpis[date < min_date] = np.nan

        # TODO check if the date difference is greater than 3 months

        if cpis.size == 1:
            return cpis.item()

        return cpis

    def calc_inflation(
        self,
        value: Union[numbers.Number, np.ndarray, pd.Series],
        original_date: Union[datetime, str],
        evaluation_date: Union[datetime, str, None] = None,
        location: Union[Location, str] = Location.AUSTRALIA,
    ):
        """
        Adjusts a value (or list of values) for inflation.

        Args:
            value (Union[numbers.Number, np.ndarray, pd.Series]): The value to be converted.
            original_date (Union[datetime, str]): The date that the value is in relation to.
            evaluation_date (Union[datetime, str], optional): The date to adjust the value to. Defaults to the current date.
            location (Union[Location, str], optional): The location for calculating the CPI.
                Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
                Default is 'Australia'.

        Returns:
            Union[float, np.ndarray]: The adjusted value.
        """
        if evaluation_date is None:
            evaluation_date = datetime.now()

        original_cpi = self.cpi_at(original_date, location=location)
        evaluation_cpi = self.cpi_at(evaluation_date, location=location)
        return value * evaluation_cpi / original_cpi

    def calc_inflation_timeseries(
        self,
        compare_date: Union[datetime, str],
        start_date: Union[datetime, str, None] = None,
        end_date: Union[datetime, str, None] = None,
        value=1,
        location: Union[Location, str] = Location.AUSTRALIA,
    ) -> pd.Series:
        compare_date = convert_date(compare_date)
        if start_date != None:
            start_date = convert_date(start_date).item()
        if end_date != None:
            end_date = convert_date(end_date).item()

        cpi_ts = self.cpi_series(location=location)[start_date:end_date].copy()
        evaluation_cpi = self.cpi_at(compare_date, location=location)
        inflation = value * (evaluation_cpi / cpi_ts)
        return inflation

    def plot_inflation_timeseries(
        self,
        compare_date: Union[datetime, str],
        start_date: Union[datetime, str, None] = None,
        end_date: Union[datetime, str, None] = None,
        value: Union[float, int] = 1,
        location: Union[Location, str] = Location.AUSTRALIA,
        **kwargs,
    ) -> plotly.graph_objects.Figure:
        """
        Plots a time series of dollar values attached to a particular date's dollar value

        Args:
            compare_date (datetime, str): Date to set relative value of the dollars too.
            start_date (datetime, str, optional): Date to set the beginning of the time series graph. Defaults to None, which starts in 1948.
            end_date (datetime, str, optional): Date to set the end of the time series graph too. Defaults to None, which will set the end date to the most recent quarter.
            value (float, int, optional): Value you in `compare_date` dollars to plot on the time series. Defaults to 1.
            location (Location, str, optional): The location for calculating the CPI.
                Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
                Default is 'Australia'.
            kwargs: (Optional(dict)): additional parameters to feed into plotly.express.line function

        Returns:
            plotly.graph_objects.Figure: line graph of inflated dollar values vs time
        """

        inflation = self.calc_inflation_timeseries(
            compare_date, start_date, end_date, value=value, location=location
        ).reset_index()
        new_col_name = f"Equivalent Dollar Value"
        if "title" not in kwargs:
            kwargs["title"] = f"The equivalent of ${value:.2f} from {str(compare_date)}"
        inflation.rename(
            columns={self.column_name(location): new_col_name},
            inplace=True,
        )
        fig = px.line(inflation, x="Date", y=new_col_name, **kwargs)
        format_fig(fig)
        return fig

    def plot_cpi_timeseries(
        self,
        start_date: Union[datetime, str, None] = None,
        end_date: Union[datetime, str, None] = None,
        locations: List[Location] = None,
        title: str = None,
        **kwargs,
    ) -> plotly.graph_objects.Figure:
        """
        Plots CPI vs time.

        Args:
            start_date (datetime, str, optional): Date to set the beginning of the time series graph. Defaults to None, which starts in 1948.
            end_date (datetime, str, optional): Date to set the end of the time series graph too. Defaults to None, which will set the end date to the most recent quarter.
            locations (List[Location], optional): The location(s) for calculating the CPI.
                Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
                Default is 'Australia'.
            title: (str, optional): The title of the figure.
            kwargs:: additional parameters to feed into plotly.express.line function.

        Returns:
            plotly.graph_objects.Figure: plot of cpi vs time
        """
        if not locations:
            locations = list(Location)

        if start_date is not None:
            start_date = convert_date(start_date).item()
        if end_date is not None:
            end_date = convert_date(end_date).item()

        df = self.latest_cpi_df
        column_map = {self.column_name(location): str(location) for location in locations}
        df = df.rename(columns=column_map)
        df = df[column_map.values()]
        df = df[start_date:end_date].copy()
        df = df.reset_index()

        fig = px.line(df, x="Date", y=list(column_map.values()), **kwargs)
        fig.update_layout(
            yaxis_title="CPI",
            legend_title="Location",
        )

        if title is None:
            location_name = locations[0] if len(locations) == 1 else "Australia"
            title = f"Consumer Price Index in {location_name} over time"

        fig.update_layout(title=title)

        if len(locations) == 1:
            fig.update_layout(
                showlegend=False,
            )
        format_fig(fig)
        return fig


_cpi = CPI()


def calc_inflation(
    value: Union[numbers.Number, np.ndarray, pd.Series],
    original_date: Union[datetime, str],
    evaluation_date: Union[datetime, str] = None,
    location: Union[Location, str] = Location.AUSTRALIA,
) -> Union[float, np.ndarray]:
    """
    Adjusts a value (or list of values) for inflation.

    Args:
        value (numbers.Number, np.ndarray, pd.Series): The value to be converted.
        original_date (datetime, str): The date that the value is in relation to.
        evaluation_date (datetime, str, optional): The date to adjust the value to. Defaults to the current date.
        location (Location, str, optional): The location for calculating the CPI.
            Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
            Default is 'Australia'.

    Returns:
        Union[float, np.ndarray]: The adjusted value.
    """
    return _cpi.calc_inflation(
        value,
        original_date=original_date,
        evaluation_date=evaluation_date,
        location=location,
    )


def plot_inflation_timeseries(
    compare_date: Union[datetime, str],
    start_date: Union[datetime, str, None] = None,
    end_date: Union[datetime, str, None] = None,
    value: Union[float, int] = 1,
    location: Union[Location, str] = Location.AUSTRALIA,
    **kwargs,
) -> plotly.graph_objects.Figure:
    """
    Plots a time series of dollar values attached to a particular date's dollar value.

    Args:
        compare_date (datetime, str): Date to set relative value of the dollars too.
        start_date (datetime, str, optional): Date to set the beginning of the time series graph. Defaults to None, which starts in 1948.
        end_date (datetime, str, optional): Date to set the end of the time series graph too. Defaults to None, which will set the end date to the most recent quarter.
        value (float, int, optional): Value you in `compare_date` dollars to plot on the time series. Defaults to 1.
        kwargs: additional parameters to feed into plotly.express.line function

    Returns:
        plotly.graph_objects.Figure: line graph of inflated dollar values vs time
    """
    return _cpi.plot_inflation_timeseries(
        compare_date, start_date=start_date, end_date=end_date, value=value, location=location, **kwargs
    )


def plot_cpi_timeseries(
    start_date: Union[datetime, str, None] = None,
    end_date: Union[datetime, str, None] = None,
    **kwargs,
) -> plotly.graph_objects.Figure:
    """
    Plots the Australian CPI vs time

    Args:
        start_date (datetime, str, optional): Date to set the beginning of the time series graph. Defaults to None, which starts in 1948.
        end_date (datetime, str, optional): Date to set the end of the time series graph too. Defaults to None, which will set the end date to the most recent quarter.
        kwargs: additional parameters to feed into plotly.express.line function.

    Returns:
        plotly.graph_objects.Figure: plot of cpi vs time
    """
    return _cpi.plot_cpi_timeseries(start_date=start_date, end_date=end_date, **kwargs)


def latest_cpi_df() -> pd.DataFrame:
    """
    Returns a pandas DataFrame with the latest CPI data from the Australian Bureau of Statistics.

    This will contain the data for each quarter going back to 1948.

    Returns:
        pandas.DataFrame: The latest dataframe with the CPI data.

    """
    return _cpi.latest_cpi_df



from pathlib import Path
import plotly.io as pio

pio.kaleido.scope.mathjax = None


def format_fig(fig):
    """Formats a plotly figure in a nicer way."""
    fig.update_layout(
        width=1200,
        height=550,
        plot_bgcolor="white",
        title_font_color="black",
        font=dict(
            family="Linux Libertine Display O",
            size=18,
            color="black",
        ),
    )
    gridcolor = "#dddddd"
    fig.update_xaxes(gridcolor=gridcolor)
    fig.update_yaxes(gridcolor=gridcolor)

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', mirror=True, ticks='outside')
    fig.update_yaxes(showline=True, linewidth=1, linecolor='black', mirror=True, ticks='outside')


def write_fig(fig, output: Path):
    """
    Writes a plotly figure to file.

    Args:
        fig (plotly figure): The figure to be written
        output (Path): The path to the output file.
            If the directory does not exist, then it is created.
            Output can be PDF, SVG, JPG, PNG or HTML based on the extension.
    """
    output.parent.mkdir(exist_ok=True, parents=True)
    if output.suffix.lower() == ".html":
        fig.write_html(output)
    else:
        fig.write_image(output)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ausdex']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'importlib-metadata<4.3',
 'kaleido==0.2.1',
 'modin>=0.10.2,<0.11.0',
 'numpy>=1.22.0,<2.0.0',
 'openpyxl>=3.0.0,<4.0.0',
 'pandas>=1.3.2,<2.0.0',
 'plotly>=5.4.0,<6.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'typer>=0.4.0,<0.5.0',
 'xlrd>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['ausdex = ausdex.main:app']}

setup_kwargs = {
    'name': 'ausdex',
    'version': '1.1.1',
    'description': 'A Python package for adjusting Australian dollars for inflation',
    'long_description': '# ausdex\n\n![pipline](https://github.com/rbturnbull/ausdex/actions/workflows/pipeline.yml/badge.svg)\n[<img src="https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/rbturnbull/49262550cc8b0fb671d46df58de213d4/raw/coverage-badge.json">](<https://rbturnbull.github.io/ausdex/coverage/>)\n[<img src="https://github.com/rbturnbull/ausdex/actions/workflows/docs.yml/badge.svg">](<https://rbturnbull.github.io/ausdex/>)\n[<img src="https://img.shields.io/badge/code%20style-black-000000.svg">](<https://github.com/psf/black>)\n[![Contributor Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.1-4baaaa.svg)](code_of_conduct.md)\n[![status](https://joss.theoj.org/papers/817baa72d2b17b535af8f421a43404b0/status.svg)](https://joss.theoj.org/papers/817baa72d2b17b535af8f421a43404b0)\n\nA Python package for adjusting Australian dollars for inflation.\n\nThe Australian Bureau of Statistics (ABS) publishes the Consumer Price Index (CPI) \nfor Australia and its capital cities which allows for adjustment of the value of Australian dollars for inflation. \n`ausdex` makes these data available with an inflation calculator in a convenient Python package with simple programmatic and command-line interfaces.\n\nABS datasets are generally housed in Microsoft Excel spreadsheets linked from the data catalogue. Working with these spreadsheets directly is cumbersome. The `ausdex` package provides an Application Programming Interface (API) for Australian CPI data that seemlessly interoperates with `NumPy` and `pandas`. It makes working with Australian dollars in Python convenient in a similar manner to the [cpi](https://github.com/palewire/cpi) Python package which adjusts US dollars for inflation.\n\nThe package is documented here: https://rbturnbull.github.io/ausdex\n\n## Installation\n\nYou can install `ausdex` from the Python Package Index (PyPI):\n\n```\npip install ausdex\n```\n\n`ausdex` requires Python 3.8 or higher.\n\nTo install ausdex for development, see the documentation for [contributing](https://rbturnbull.github.io/ausdex/contributing.html).\n\n## Command Line Usage\n\nAdjust single values using the command line interface:\n```\nausdex inflation VALUE ORIGINAL_DATE\n```\nThis adjust the value from the original date to the equivalent for the most recent quarter.\n\nFor example, to adjust $26 from July 21, 1991 to the latest quarter run:\n```\n$ ausdex inflation 26 "July 21 1991" \n$ 52.35\n```\n\nTo choose a different date for evaluation use the `--evaluation-date` option. This adjusts the value to dollars in the quarter corresponding to that date. For example, this command adjusts $26 from July 1991 to dollars in September 1999:\n```\n$ ausdex inflation 26 "July 21 1991"  --evaluation-date "Sep 1999"\n$ 30.27\n```\n\nBy default, `ausdex` uses the CPI for Australia in general but you can calculate the inflation for specific capital cities with the `--location` argument:\n```\n$ ausdex inflation 26 "July 21 1991"  --evaluation-date "Sep 1999" --location sydney\n$ 30.59\n```\n\nLocation options are: \'Australia\', \'Sydney\', \'Melbourne\', \'Brisbane\', \'Adelaide\', \'Perth\', \'Hobart\', \'Darwin\', and \'Canberra\'.\n\n\n## Module Usage\n\n```\n>>> import ausdex\n>>> ausdex.calc_inflation(26, "July 21 1991")\n52.35254237288135\n>>> ausdex.calc_inflation(26, "July 21 1991", evaluation_date="Sep 1999")\n30.27457627118644\n>>> ausdex.calc_inflation(26, "July 21 1991", evaluation_date="Sep 1999", location="sydney")\n30.59083191850594\n```\nThe dates can be as strings or Python datetime objects.\n\nThe values, the dates and the evaluation dates can be vectors by using NumPy arrays or Pandas Series. e.g.\n```\n>>> df = pd.DataFrame(data=[ [26, "July 21 1991"],[25,"Oct 1989"]], columns=["value","date"] )\n>>> df[\'adjusted\'] = ausdex.calc_inflation(df.value, df.date)\n>>> df\n   value          date   adjusted\n0     26  July 21 1991  52.352542\n1     25      Oct 1989  54.797048\n```\n\n## Dataset and Validation\n\nThe Consumer Price Index dataset is taken from the [Australian Bureau of Statistics](https://www.abs.gov.au/statistics/economy/price-indexes-and-inflation/consumer-price-index-australia). It uses the nation-wide CPI value. The validation examples in the tests are taken from the [Australian Reserve Bank\'s inflation calculator](https://www.rba.gov.au/calculator/). This will automatically update each quarter as the new datasets are released.\n\nThe CPI data goes back to 1948. Using dates before this will result in a NaN.\n\nTo access the raw CPI data as a pandas DataFrame, use this function:\n```\ndf = ausdex.latest_cpi_df()\n```\n\nThe Excel spreadsheet for this is stored in the user\'s cache directory. \nIf you wish to download this Excel file to a specific location, use this function:\n```\nausdex.files.cached_download_cpi(local_path="cpi-data.xlsx")\n```\n\nFor more infomation about the methods to download data from the ABS, see the [API specification](https://rbturnbull.github.io/ausdex/reference.html).\n\n## Contributing\n\nSee the guidelines for contributing and our code of conduct in the [documentation](https://rbturnbull.github.io/ausdex/contributing.html).\n\n## License and Disclaimer\n\n`ausdex` is released under the Apache 2.0 license.\n\nWhile every effort has been made by the authors of this package to ensure that the data and calculations used to produce the results are accurate, as is stated in the license, we accept no liability or responsibility for the accuracy or completeness of the calculations. \nWe recommend that users exercise their own care and judgment with respect to the use of this package.\n \n## Credits\n\n`ausdex` was written by [Dr Robert Turnbull](https://findanexpert.unimelb.edu.au/profile/877006-robert-turnbull) and [Dr Jonathan Garber](https://findanexpert.unimelb.edu.au/profile/787135-jonathan-garber) from the [Melbourne Data Analytics Platform](https://mdap.unimelb.edu.au/).\n\nPlease cite from the article when it is released. Details to come soon.\n\n## Acknowledgements\n\nThis project came about through a research collaboration with [Dr Vidal Paton-Cole](https://findanexpert.unimelb.edu.au/profile/234417-vidal-paton-cole) and [Prof Robert Crawford](https://findanexpert.unimelb.edu.au/profile/174016-robert-crawford). We acknowledge the support of our colleagues at the Melbourne Data Analytics Platform: [Dr Aleksandra Michalewicz](https://findanexpert.unimelb.edu.au/profile/27349-aleks-michalewicz) and [Dr Emily Fitzgerald](https://findanexpert.unimelb.edu.au/profile/196181-emily-fitzgerald).\n',
    'author': 'Robert Turnbull',
    'author_email': 'robert.turnbull@unimelb.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rbturnbull/ausdex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import List, Union
import webbrowser
import typer

import importlib_metadata as lib_metadata
from typing import Optional
import subprocess

from ausdex import Location, calc_inflation

from . import viz

app = typer.Typer()


def version_callback(value: bool):
    if value:
        version = lib_metadata.version("ausdex")
        typer.echo(version)
        raise typer.Exit()


@app.command()
def repo():
    """
    Opens the repository in a web browser.
    """
    typer.launch("https://github.com/rbturnbull/ausdex")


@app.command()
def docs(live: bool = True):
    """Builds the documentation.

    Args:
        live (bool, optional): Whether or not to use sphinx-autobuild to automatically build the documentation as files are edited. Defaults to True.
    """
    root_dir = Path(__file__).parent.parent.resolve()
    docs_dir = root_dir / "docs"
    docs_build_dir = docs_dir / "_build/html"

    if live:
        command = f"sphinx-autobuild {docs_dir} {docs_build_dir} --open-browser"
    else:
        command = f"sphinx-build -E -b html {docs_dir} {docs_build_dir}"

    subprocess.run(command, shell=True)

    if not live:
        index_page = docs_build_dir / "index.html"
        print(f"Open the index page at {index_page}")
        webbrowser.open_new("file://" + str(index_page))


@app.command()
def inflation(
    value: float = typer.Argument(..., help="The dollar value to be converted."),
    original_date: str = typer.Argument(..., help="The date that the value is in relation to."),
    evaluation_date: str = typer.Option(None, help="The date to adjust the value to. Defaults to the current date."),
    location: Location = typer.Option(
        Location.AUSTRALIA, case_sensitive=False, help="The location for calculating the CPI."
    ),
):
    """
    Adjusts Australian dollars for inflation.

    Prints output to stdout.

    Args:
        value (float): The dollar value to be converted.
        original_date (str): The date that the value is in relation to.
        evaluation_date (str, optional): The date to adjust the value to. Defaults to the current date.
        location (Location, optional): The location for calculating the CPI.
            Options are 'Australia', 'Sydney', 'Melbourne', 'Brisbane', 'Adelaide', 'Perth', 'Hobart', 'Darwin', and 'Canberra'.
            Default is 'Australia'.
    """

    result = calc_inflation(
        value=value, original_date=original_date, evaluation_date=evaluation_date, location=location
    )
    typer.echo(f"{result:.2f}")


@app.command()
def plot_inflation(
    compare_date: str = typer.Argument(..., help="Date to set relative value of the dollars too."),
    show: bool = typer.Option(True, help="Whether or not to show the figure in a browser."),
    output: Path = typer.Option(
        None,
        help="The path to where the figure will be saved. Output can be PDF, JPG, PNG or HTML based on the extension.",
    ),
    start_date: str = typer.Option(
        None, help="Date to set the beginning of the time series graph. Defaults to None, which starts in 1948."
    ),
    end_date: str = typer.Option(
        None,
        help="Date to set the end of the time series graph too. If empty, then the end date to the most recent quarter.",
    ),
    value: float = typer.Option(
        1.0, help="Value you in `compare_date` dollars to plot on the time series. Defaults to 1."
    ),
    location: Location = typer.Option(
        Location.AUSTRALIA,
        help="The location for calculating the CPI.",
        case_sensitive=False,
    ),
):
    """
    Plots a time series of dollar values attached to a particular date's dollar value.

    Args:
        compare_date (str): Date to set relative value of the dollars too.
        out (Path): The path to where the figure will be saved. Output can be PDF, JPG, PNG or HTML based on the extension.
        start_date (str, optional): Date to set the beginning of the time series graph. Defaults to None, which starts in 1948.
        end_date (str, optional): Date to set the end of the time series graph too. Defaults to None, which will set the end date to the most recent quarter.
        value (float, optional): Value you in `compare_date` dollars to plot on the time series. Defaults to 1.
    """
    from ausdex.inflation import plot_inflation_timeseries

    fig = plot_inflation_timeseries(
        compare_date=compare_date, start_date=start_date, end_date=end_date, value=value, location=location
    )
    if output:
        print(f"Writing figure to '{output}'.")
        viz.write_fig(fig, output)
    if show:
        fig.show()


@app.command()
def plot_cpi(
    show: bool = typer.Option(True, help="Whether or not to show the figure in a browser."),
    output: Path = typer.Option(
        None,
        help="The path to where the figure will be saved. Output can be PDF, JPG, PNG or HTML based on the extension.",
    ),
    start_date: str = typer.Option(
        None, help="Date to set the beginning of the time series graph. Defaults to None, which starts in 1948."
    ),
    end_date: str = typer.Option(
        None,
        help="Date to set the end of the time series graph too. If empty, then the end date to the most recent quarter.",
    ),
    location: List[Location] = typer.Option(
        None,
        help="The location for calculating the CPI.",
        case_sensitive=False,
    ),
    title: str = typer.Option(None, help="A custom title of the plot."),
):
    """
    Plot the Australian CPI over time.

    Args:
        output (Path): The path to where the figure will be saved. Output can be PDF, SVG, JPG, PNG or HTML based on the extension.
        start_date (str, optional): Date to set the beginning of the time series graph. If empty, it defaults to 1948.
        end_date (str, optional): Date to set the end of the time series graph too. If empty, then the end date to the most recent quarter.
        location (List[location]): The location for calculating the CPI.
        title (str, optional): A custom title of the plot.
    """
    from ausdex.inflation import plot_cpi_timeseries

    fig = plot_cpi_timeseries(start_date=start_date, end_date=end_date, locations=location, title=title)
    if output:
        print(f"Writing figure to '{output}'.")
        viz.write_fig(fig, output)
    if show:
        fig.show()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(None, "--version", "-v", callback=version_callback, is_eager=True),
):
    """Adjusts Australian dollars for inflation."""
    pass



import sys
from pathlib import Path
from appdirs import user_cache_dir
from typing import Union
import urllib.request
from datetime import datetime, timedelta


ACCEPTED_QUARTERS = ("mar", "jun", "sep", "dec")


class DownloadError(Exception):
    pass


def get_cached_path(filename: str) -> Path:
    """
    Returns a path in the ausdex directory in the user's cache.

    File may or may not exist.
    """
    cache_dir = Path(user_cache_dir("ausdex"))
    cache_dir.mkdir(exist_ok=True, parents=True)
    return cache_dir / filename


def cached_download(url: str, local_path: Union[str, Path], force: bool = False) -> None:
    """
    Downloads a file if a local file does not already exist.

    Args:
        url (str): The url of the file to download.
        local_path (str, Path): The local path of where the file should be.
            If this file isn't there or the file size is zero then this function downloads it to this location.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
            Default False.

    Raises:
        DownloadError: Raises an exception if it cannot download the file.
        IOError: Raises an exception if the file does not exist or is empty after downloading.
    """
    local_path = Path(local_path)
    if (not local_path.exists() or local_path.stat().st_size == 0) or force:
        try:
            print(f"Downloading {url} to {local_path}")
            urllib.request.urlretrieve(url, local_path)
        except:
            raise DownloadError(f"Error downloading {url}")

    if not local_path.exists() or local_path.stat().st_size == 0:
        raise IOError(f"Error reading {local_path}")


def cached_download_abs(
    id: str,
    quarter: str,
    year: Union[int, str],
    extension: str,
    local_path: Union[Path, str, None] = None,
    force: bool = False,
) -> Path:
    """
    Downloads a file from the ABS if a local file does not already exist.

    Args:
        id (str): The ABS id for the datafile. For Australian Consumer Price Index the ID is 640101.
        quarter (str): The quarter of the file in question. One of "mar", "jun", "sep", or "dec".
        year (str, int): The year for the file in question.
        extension (str): The extension of the file in question.
        local_path (Path, str, optional): The path to where the file should be downloaded.
            If None, then it is downloaded in the user's cache directory.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
        local_path (str, Path): The local path of where the file should be.
            If this file isn't there or the file size is zero then this function downloads it to this location.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.

    Raises:
        ValueError: If the value for `quarter` cannot be understood.
        DownloadError: Raises an exception if it cannot download the file.
        IOError: Raises an exception if the file does not exist or is empty after downloading.

    Returns:
        Path: The path to the downloaded ABS datafile.
    """
    quarter = quarter.lower()[:3]
    if quarter not in ACCEPTED_QUARTERS:
        raise ValueError(f"Cannot understand quarter {quarter}.")

    url = f"https://www.abs.gov.au/statistics/economy/price-indexes-and-inflation/consumer-price-index-australia/{quarter}-{year}/{id}.{extension}"
    local_path = local_path or get_cached_path(f"{id}-{quarter}-{year}.{extension}")
    local_path = Path(local_path)
    cached_download(url, local_path, force=force)
    return local_path


def cached_download_abs_excel(
    id: str, quarter: str, year: Union[str, int], local_path: Union[Path, str, None] = None, force: bool = False
) -> Path:
    """
    Gets am Excel file from the Australian Burau of Statistics.

    First it tries the extension `xlsx` and then falls back to `xls`.

    Args:
        id (str): The ABS id for the datafile. For Australian Consumer Price Index the ID is 640101.
        quarter (str): The quarter of the file in question. One of "mar", "jun", "sep", or "dec".
        year (str, int): The year for the file in question.
        local_path (Path, str, optional): The path to where the file should be downloaded.
            If None, then it is downloaded in the user's cache directory.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
            Default False.

    Raises:
        ValueError: Raises this error if the quarter cannot be understood.

    Returns:
        Path: The path to the cached ABS datafile
    """
    try:
        local_path = cached_download_abs(
            quarter=quarter, year=year, id=id, extension="xlsx", local_path=local_path, force=force
        )
    except (DownloadError, IOError):
        local_path = cached_download_abs(
            quarter=quarter, year=year, id=id, extension="xls", local_path=local_path, force=force
        )

    return local_path


def cached_download_abs_excel_by_date(
    id: str, date: Union[datetime, None] = None, local_path: Union[Path, str, None] = None, force: bool = False
) -> Path:
    """
    Gets a datafile from the Australian Burau of Statistics before a specific date.

    Args:
        id (str): The ABS id for the datafile. For Australian Consumer Price Index (CPI) the ID is 640101.
        date (datetime, optional): The date before which the CPI data should be valid.
            If not provided, then it uses today's date download get the latest file.
        local_path (Path, str, optional): The path to where the file should be downloaded.
            If None, then it is downloaded in the user's cache directory.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
            Default False.

    Returns:
        Path: The path to the cached ABS datafile.
    """
    date = date or datetime.now()
    file = None
    while file is None and date > datetime(1948, 1, 1):
        try:
            year = date.year
            quarter_index = (date.month - 3) // 3
            if quarter_index == -1:
                quarter_index = 3
                year -= 1
            quarter = ACCEPTED_QUARTERS[quarter_index]

            file = cached_download_abs_excel(id, quarter, year, local_path=local_path, force=force)
        except:
            print(f"CPI data for {date} not available.", file=sys.stderr)

        date -= timedelta(days=89)  # go back approximately a quarter

    return file


def cached_download_cpi(
    *, date: Union[datetime, None] = None, local_path: Union[Path, str, None] = None, force: bool = False
) -> Path:
    """
    Returns the path to the latest cached file with the Australian Consumer Price Index (CPI) data.

    It downloads the file if it does not exist already. The ABS id of this file is "640101".

    Args:
        date (datetime, optional): The date before which the CPI data should be valid.
            If not provided, then it uses today's date download get the latest file.
        local_path (Path, str, optional): The path to where the file should be downloaded.
            If None, then it is downloaded in the user's cache directory.
        force (bool): Whether or not the file should be forced to download again even if present in the local path.
            Default False.

    Returns:
        Path: The path to the cached datafile.
    """
    CPI_FILE_ID = "640101"
    return cached_download_abs_excel_by_date(id=CPI_FILE_ID, date=date, local_path=local_path, force=force)
